package middleware

import (
	"bytes"
	"gitlab.wanzhuangkj.com/tush/xpkg/httpcli/entity"
	"io"
	"net/http"
	"time"

	ctxUtils "gitlab.wanzhuangkj.com/tush/xpkg/gin/ctxUtils"
	"gitlab.wanzhuangkj.com/tush/xpkg/ips"

	"github.com/gin-gonic/gin"
	"go.uber.org/zap"
)

var (
	// Print body max length
	defaultMaxLength = 300

	// default zap log
	defaultLogger, _ = zap.NewProduction()

	// Ignore route list
	defaultIgnoreRoutes = map[string]struct{}{
		"/ping":   {},
		"/pong":   {},
		"/health": {},
	}

	emptyBody   = []byte("")
	contentMark = []byte(" ...... ")
)

// Option set the gin logger options.
type Option func(*options)

func defaultOptions() *options {
	return &options{
		maxLength:    defaultMaxLength,
		log:          defaultLogger,
		ignoreRoutes: defaultIgnoreRoutes,
		traceIDFrom:  0,
	}
}

type options struct {
	maxLength    int
	appName      string
	log          *zap.Logger
	ignoreRoutes map[string]struct{}
	traceIDFrom  int // 0: ignore, 1: from context, 2: from header
}

func (o *options) apply(opts ...Option) {
	for _, opt := range opts {
		opt(o)
	}
}

// WithMaxLen logger content max length
func WithMaxLen(maxLen int) Option {
	return func(o *options) {
		if maxLen < len(contentMark) {
			panic("maxLen should be greater than or equal to 8")
		}
		o.maxLength = maxLen
	}
}

// WithLog set log
func WithLog(log *zap.Logger) Option {
	return func(o *options) {
		if log != nil {
			o.log = log
		}
	}
}
func WithAppName(appName string) Option {
	return func(o *options) {
		o.appName = appName
	}
}

// WithIgnoreRoutes no logger content routes
func WithIgnoreRoutes(routes ...string) Option {
	return func(o *options) {
		for _, route := range routes {
			o.ignoreRoutes[route] = struct{}{}
		}
	}
}

// WithTraceIDFromContext name is field in context, default value is request_id
func WithTraceIDFromContext() Option {
	return func(o *options) {
		o.traceIDFrom = 1
	}
}

// WithTraceIDFromHeader name is field in header, default value is X-Request-Id
func WithTraceIDFromHeader() Option {
	return func(o *options) {
		o.traceIDFrom = 2
	}
}

// ------------------------------------------------------------------------------------------

type bodyLogWriter struct {
	gin.ResponseWriter
	body *bytes.Buffer
}

func (w bodyLogWriter) Write(b []byte) (int, error) {
	w.body.Write(b)
	return w.ResponseWriter.Write(b)
}

// If there is sensitive information in the body, you can use WithIgnoreRoutes set the route to ignore logging
func getResponseBody(buf *bytes.Buffer, maxLen int) []byte {
	l := buf.Len()
	if l == 0 {
		return []byte("")
	} else if l > maxLen {
		l = maxLen
	}

	body := make([]byte, l)
	n, _ := buf.Read(body)
	if n == 0 {
		return emptyBody
	} else if n < maxLen {
		if isLastByteNewline(body) {
			return body[:n-1]
		}
		return body[:n]
	}
	return append(body[:maxLen-len(contentMark)], contentMark...)
}

func isLastByteNewline(data []byte) bool {
	if len(data) == 0 {
		return false
	}
	return data[len(data)-1] == '\n'
}

// If there is sensitive information in the body, you can use WithIgnoreRoutes set the route to ignore logging
func getRequestBody(buf *bytes.Buffer, maxLen int) []byte {
	l := buf.Len()
	if l == 0 {
		return []byte("")
	} else if l < maxLen {
		return buf.Bytes()
	}

	body := make([]byte, maxLen)
	copy(body, buf.Bytes())
	return append(body[:maxLen-len(contentMark)], contentMark...)
}

// Logging print request and response info
func Logging(opts ...Option) gin.HandlerFunc {
	o := defaultOptions()
	o.apply(opts...)

	return func(c *gin.Context) {
		start := time.Now()

		// ignore printing of the specified route
		if _, ok := o.ignoreRoutes[c.Request.URL.Path]; ok {
			c.Next()
			return
		}

		// print input information before processing
		buf := &bytes.Buffer{}
		_, _ = buf.ReadFrom(c.Request.Body)

		fields := []zap.Field{
			zap.String("method", c.Request.Method),
			zap.String("url", c.Request.URL.String()),
			zap.Any("headers", c.Request.Header),
		}
		if c.Request.Method == http.MethodPost || c.Request.Method == http.MethodPut || c.Request.Method == http.MethodPatch || c.Request.Method == http.MethodDelete {
			fields = append(fields,
				zap.Int("size", buf.Len()),
				zap.ByteString("body", getRequestBody(buf, o.maxLength)),
			)
		}

		entity.SetCopyReq(c, buf)

		c.Set(ctxUtils.KeyAppName, o.appName)

		reqID := ""
		if o.traceIDFrom == 1 {
			if v, isExist := c.Get(ctxUtils.ContextTraceIDKey); isExist {
				if requestID, ok := v.(string); ok {
					reqID = requestID
					fields = append(fields, zap.String(ctxUtils.ContextTraceIDKey, reqID))
				}
			}
		} else if o.traceIDFrom == 2 {
			reqID = c.Request.Header.Get(ctxUtils.HeaderXRequestIDKey)
			fields = append(fields, zap.String(ctxUtils.ContextTraceIDKey, reqID))
		}
		o.log.Info("<<<<<<<<<req", fields...)

		c.Request.Body = io.NopCloser(buf)

		//replace writer
		newWriter := &bodyLogWriter{body: &bytes.Buffer{}, ResponseWriter: c.Writer}
		c.Writer = newWriter

		ip := ips.GetClientIP(c)
		c.Set(ctxUtils.KeyClientIP, ip)

		// processing requests
		c.Next()

		//print return message after processing
		fields = []zap.Field{
			zap.Int("code", c.Writer.Status()),
			zap.String("method", c.Request.Method),
			zap.String("url", c.Request.URL.Path),
			zap.String("client_ip", ip),
			zap.String("cost", time.Since(start).String()),
			zap.Int("size", newWriter.body.Len()),
			zap.Any("header", c.Writer.Header()),
			zap.ByteString("body", getResponseBody(newWriter.body, o.maxLength)),
		}
		if reqID != "" {
			fields = append(fields, zap.String(ctxUtils.ContextTraceIDKey, reqID))
		}

		o.log.Info(">>>>>>>>>rsp", fields...)

	}
}

// SimpleLog print response info
func SimpleLog(opts ...Option) gin.HandlerFunc {
	o := defaultOptions()
	o.apply(opts...)

	return func(c *gin.Context) {
		start := time.Now()

		// ignore printing of the specified route
		if _, ok := o.ignoreRoutes[c.Request.URL.Path]; ok {
			c.Next()
			return
		}

		reqID := ""
		if o.traceIDFrom == 1 {
			if v, isExist := c.Get(ctxUtils.ContextTraceIDKey); isExist {
				if requestID, ok := v.(string); ok {
					reqID = requestID
				}
			}
		} else if o.traceIDFrom == 2 {
			reqID = c.Request.Header.Get(ctxUtils.HeaderXRequestIDKey)
		}

		// processing requests
		c.Next()

		// print return message after processing
		fields := []zap.Field{
			zap.Int("code", c.Writer.Status()),
			zap.String("method", c.Request.Method),
			zap.String("url", c.Request.URL.String()),
			zap.String("cost", time.Since(start).String()),
			zap.Int("size", c.Writer.Size()),
		}
		if reqID != "" {
			fields = append(fields, zap.String(ctxUtils.ContextTraceIDKey, reqID))
		}
		o.log.Info("Gin msg", fields...)
	}
}
